/*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*
 * volume.c -
 *    The following file is taken from xcdmix :
 *       Copyright (C) 1994
 *           Olav Woelfelschneider (wosch@rbg.informatik.th-darmstadt.de)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
\*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*/

#include  <unistd.h>
#include  <stdio.h>
#include  <fcntl.h>
#include  <linux/soundcard.h>
#include  <sys/ioctl.h>

#include  "volume.h"


/*=========================================================================
 * Static constants
\*=========================================================================*/
#define MIXERDEVICE "/dev/mixer"


/*=========================================================================
 * Start of Code
\*=========================================================================*/

static int              mixer = -1;

int    mix_open ()
{
    if (mixer < 0)
	mixer = open (MIXERDEVICE, O_RDONLY);
    return mixer;
}

void     mix_close ()
{
    if (mixer >= 0) {
	close (mixer);
	mixer = -1;
    }
}

int     mix_set_volume (int req, int *left, int *right)
{
    int                     res,
                            n;

    if (mixer < 0)
	mix_open ();
    if (mixer >= 0) {
	if ((*left >= 0) && (*right >= 0)) {
	    n = (*left & 0x7F) + ((*right & 0x7F) << 8);
	    res = ioctl (mixer, MIXER_WRITE (req), &n);
	} else {
	    res = 0;
	}
	if (res >= 0) {
	    res = ioctl (mixer, MIXER_READ (req), &n);
	    *left = n & 0x7F;
	    *right = (n >> 8) & 0x7F;
	}
    } else
	res = -1;
    if (res < 0)
	*left = *right = 0;
    return res;
}

int   mix_get_volume (int req, int *left, int *right)
{
    int                     res,
                            n;

    if (mixer < 0)
	mix_open ();
    if (mixer >= 0) {
	res = ioctl (mixer, MIXER_READ (req), &n);
	*left = n & 0x7F;
	*right = (n >> 8) & 0x7F;
    } else
	res = -1;
    if (res < 0)
	*left = *right = 0;
    return res;
}


/*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*
 *     
 * volume.c - End of File
\*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*/
